using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Linq;

using UnityEngine;
using UnityEngine.Internal;
using UnityEngine.UI;

using MVR;
using MeshVR;
using Battlehub.UIControls;

namespace Stopper {
    public static class MVRScriptExtensions {

        public delegate void TextureLoadedCallback(Texture2D t);

        public static void LoadTexture(this MVRScript script, string url, TextureLoadedCallback callback)
		{	
			if (ImageLoaderThreaded.singleton == null)
			{
				return;
			}
			if (string.IsNullOrEmpty(url))
			{	
				return;
			}
			ImageLoaderThreaded.QueuedImage queuedImage = new ImageLoaderThreaded.QueuedImage();
			queuedImage.imgPath = url;	
            queuedImage.forceReload = true;
			queuedImage.createMipMaps = false;
			queuedImage.isNormalMap = false;
			queuedImage.isThumbnail = false;
			queuedImage.linear = true;
			queuedImage.createAlphaFromGrayscale = false;
			queuedImage.compress = false;
			queuedImage.skipCache = true;
            queuedImage.invert = false;

            
			queuedImage.callback = (ImageLoaderThreaded.QueuedImage image) => {
                callback(image.tex);
			};
			ImageLoaderThreaded.singleton.QueueImage(queuedImage);			
		}

        /*
            UI
        */
        public static UIDynamicSlider AddSlider(this MVRScript script, 
                        string name, 
                        Component parent,
                        float defaultVal, 
                        float min, float max, 
                        JSONStorableFloat.SetFloatCallback onChange = null, 
                        bool rangeAdjust = true,
                        bool wholeNumbers = false, 
                        string valueFormat = "F2") {
            JSONStorableFloat storableFloat = new JSONStorableFloat(name, defaultVal, min, max, true, true);
            storableFloat.storeType = JSONStorableParam.StoreType.Full;

            return script.AddSlider(storableFloat, parent, onChange, rangeAdjust, wholeNumbers, valueFormat);
        }

        public static UIDynamicSlider AddSlider(this MVRScript script, 
                                        JSONStorableFloat storableFloat, 
                                        Component parent,
                                        JSONStorableFloat.SetFloatCallback onChange = null, 
                                        bool rangeAdjust = true,
                                        bool wholeNumbers = false, 
                                        string valueFormat = "F2" ) {
            
            UIDynamicSlider uiSlider = script.CreateSlider(storableFloat, false);
            uiSlider.rangeAdjustEnabled = rangeAdjust;
            uiSlider.slider.wholeNumbers = wholeNumbers;
            script.RegisterFloat(storableFloat);
            uiSlider.valueFormat = valueFormat;
            if (onChange != null) {
                storableFloat.setCallbackFunction = onChange;
            }
            
            uiSlider.gameObject.transform.SetParent(parent.gameObject.transform);

            return uiSlider;
        }

        public static T AddToPanel<T>(this MVRScript script, T element, LayoutPanel panel) where T : UIDynamic {
            if(element == null) {
                return null;
            }
            element.transform.SetParent(panel.transform, false);
            return element;
        }

        public static T AddUIElement<T>(this MVRScript script, string name, Component parent, bool horizontal = false) where T : UIElement {
            return AddUIElement<T>(script, name, parent.gameObject);
        }

        public static T AddUIElement<T>(this MVRScript script, string name, GameObject parent) where T : UIElement {
            var panel = CreateUIElement<T>(name, parent);
            return panel;
        }

        public static TextBox AddTextBox(this MVRScript script, string text, Component parent, TextAnchor alignment = TextAnchor.MiddleCenter, FontStyle fontStyle = FontStyle.Normal) {
            return AddTextBox(script, text, parent.gameObject, alignment, fontStyle);
        }

        public static TextBox AddTextBox(this MVRScript script, string text, GameObject parent, TextAnchor alignment = TextAnchor.MiddleCenter, FontStyle fontStyle = FontStyle.Normal) {
            TextBox textBox = AddUIElement<TextBox>(script, "TextBox", parent);
            
            textBox.UIText.text = text;
            textBox.UIText.alignment = alignment;
            textBox.UIText.fontStyle = fontStyle;

            return textBox;
        }

        public static KeybindUI AddKeybindUI(this MVRScript script, string text, JSONStorableString storable, GameObject parent) {
            KeybindUI keybind = AddUIElement<KeybindUI>(script, "KeyBind", parent);
            keybind.ratio = 0.6f;
            keybind.label.UIText.text = text;
            
            UIDynamicButton button = script.CreateButton("");
            keybind.button = button;
            AddToPanel(script, button, keybind.rightPanel);

            keybind.storable = storable;

            keybind.height = 20;
            keybind.background.color = new Color(0f,0f,0f,0.2f);

            return keybind;
        }

        public static T CreateUIElement<T>(string name, GameObject parent) where T : UIElement {
            GameObject panel = new GameObject(name);
            panel.transform.SetParent(parent.transform,false);
            
            T layoutPanel = panel.AddComponent<T>();

            return layoutPanel;
        }

        public class UIElement : MonoBehaviour {
            public RectTransform rect;
           
            public UIElement() {
                rect = this.gameObject.AddComponent<RectTransform>();
           
                rect.anchorMin = new Vector2(0f,1f);
                rect.anchorMax = new Vector2(1f,1f);

                rect.offsetMin = new Vector2(0f,0f);
                rect.offsetMax = new Vector2(0f,0f);

                rect.pivot = new Vector2(0f,1f);

                rect.anchoredPosition = new Vector2(0f,0f);
                rect.sizeDelta = new Vector2(0f,0f);
            }
        }

        public class LayoutPanel : UIElement {
            public Image background;
            public LayoutElement layoutElement;

            public ContentSizeFitter fitter;

            public LayoutPanel() : base() {
                layoutElement = this.gameObject.AddComponent<LayoutElement>();
                background = this.gameObject.AddComponent<Image>();
                background.color = new Color(0f, 0f, 0f, 0.0f);
                background.raycastTarget = false;

                fitter = this.gameObject.AddComponent<ContentSizeFitter>();
            }
        }

        public class VerticalLayoutPanel : LayoutPanel {
            public VerticalLayoutGroup layoutGroup;

            public VerticalLayoutPanel() : base() {
                fitter.horizontalFit = ContentSizeFitter.FitMode.Unconstrained;
                fitter.verticalFit = ContentSizeFitter.FitMode.PreferredSize;

                layoutGroup = this.gameObject.AddComponent<VerticalLayoutGroup>();
                
                layoutGroup.padding.left = 5;
                layoutGroup.padding.right = 5;
                layoutGroup.padding.top = 5;
                layoutGroup.padding.bottom = 5;

                layoutGroup.spacing = 10f;
            }
        }

        public class HorizontalLayoutPanel : LayoutPanel {
            public HorizontalLayoutGroup layoutGroup;

            public HorizontalLayoutPanel() : base() {
                layoutGroup = this.gameObject.AddComponent<HorizontalLayoutGroup>();

                layoutGroup.padding.left = 5;
                layoutGroup.padding.right = 5;
                layoutGroup.padding.top = 5;
                layoutGroup.padding.bottom = 5;

                layoutGroup.spacing = 10f;
            }
        }

        /*
            Layout group that takes the size of all children, but does not 
            resize them
        */
        public class MaxSizeLayoutPanel : LayoutPanel {
            public LayoutMaxSize layoutGroup;

            public MaxSizeLayoutPanel() : base() {
                layoutGroup = this.gameObject.AddComponent<LayoutMaxSize>();
                layoutGroup.adjustWidth = true;
                layoutGroup.adjustHeight = true;

                layoutGroup.padding.left = 5;
                layoutGroup.padding.right = 5;
                layoutGroup.padding.top = 5;
                layoutGroup.padding.bottom = 5;
            }
        }

        public class VerticalSplitLayoutPanel : MaxSizeLayoutPanel {
            public VerticalLayoutPanel leftPanel;
            public VerticalLayoutPanel rightPanel;

            private float _ratio = 0.5f;

            public VerticalSplitLayoutPanel() : base() {
                leftPanel = CreateUIElement<VerticalLayoutPanel>($"{this.gameObject.name}leftPanel", this.gameObject);
                rightPanel = CreateUIElement<VerticalLayoutPanel>($"{this.gameObject.name}rightPanel", this.gameObject);
                
                SyncRatio();
              
            }

            private void SyncRatio() {
                leftPanel.rect.anchorMin = new Vector2(0f,1f);
                leftPanel.rect.anchorMax = new Vector2(_ratio,1f);

                rightPanel.rect.anchorMin = new Vector2(_ratio,1f);
                rightPanel.rect.anchorMax = new Vector2(1f,1f);
            }

            public float ratio {
                get {
                    return _ratio;
                }
                set {
                    _ratio = value;
                    SyncRatio();
                }
            }
        }

        public class TextBox : VerticalLayoutPanel  {
            public LayoutPanel textPanel;
            public UIElement textElement;
            public Text UIText;
            
            public TextBox() : base() {
                textElement = CreateUIElement<UIElement>("Text", this.gameObject);
                UIText = textElement.gameObject.AddComponent<Text>();
                
                UIText.font = Resources.GetBuiltinResource(typeof(Font), "Arial.ttf") as Font;
                UIText.color = Color.black;
                UIText.fontSize = 30;
                UIText.resizeTextForBestFit = true;
                UIText.resizeTextMinSize = 10;
                UIText.resizeTextMaxSize = 30;
                UIText.alignment = TextAnchor.MiddleCenter;
                
                textElement.rect.anchorMin = new Vector2(0f, 1f);
                textElement.rect.anchorMax = new Vector2(1f, 1f);
            }

            public string text {
                get {
                    return UIText.text;
                } set {
                    UIText.text = value;
                }
            }
        }

        public class KeybindUI : VerticalSplitLayoutPanel  {
            public TextBox label;
            public List<KeyCode> bindings = new List<KeyCode>();

            private UIDynamicButton _button;
            private float _height;
            private bool _capturing;
            private JSONStorableString _storable;


            public KeybindUI() : base() {
                label = CreateUIElement<TextBox>("Label", this.leftPanel.gameObject);
            }

            
            public void Update() {
                if(_capturing) {
                    foreach(KeyCode code in Enum.GetValues(typeof(KeyCode))) {
                        if(Input.GetKey(code) && !bindings.Contains(code)) {
                            bindings.Add(code);
                            SyncButtonText();
                        }
                    }
                    if(!Input.anyKey && bindings.Count > 0) {
                        _capturing = false;
                        SuperController.singleton.disableInternalKeyBindings = false;
                        SuperController.singleton.disableInternalNavigationKeyBindings = false;
                        SyncButtonText();
                    }
                }
            }

            private void SetUpButton() {
                if(_button != null) {

                    _button.buttonText.resizeTextForBestFit = true;
                    _button.buttonText.resizeTextMinSize = 10;
                    _button.buttonText.resizeTextMaxSize = 30;
                    
                    _button.button.onClick.AddListener(() => {
                        _capturing = true;
                        SuperController.singleton.disableInternalKeyBindings = true;
                        SuperController.singleton.disableInternalNavigationKeyBindings = true;
                        bindings.Clear();
                        SyncButtonText();
                    });

                    SyncButtonText();
                }
            }

            private void SyncButtonText() {
                if(_button != null) {
                    if(_capturing && bindings.Count == 0) {
                        _button.label = "ESC to clear";
                    } else if(bindings.Count == 0) {
                        _button.label = "N/A";
                    } else if(storable != null) {
                        _button.label = storable.val;
                    }
                }
            }

            private void SyncBindingsFromString(string input) {
                foreach(string bindString in input.Split('+')) {
                    try {
                        KeyCode binding = (KeyCode)Enum.Parse(typeof(KeyCode), bindString);
                        if(!bindings.Contains(binding)) {
                            bindings.Add(binding);
                        }
                    } catch (Exception e) {
                        //invalid keybind
                    }
                }
            }

            public JSONStorableString storable {
                get {
                    return _storable;
                }
                set {
                    if(value != null) {
                        _storable = value;
                        SyncBindingsFromString(value.val);
                        SyncButtonText();
                        _storable.setCallbackFunction += (val) => {
                            SyncBindingsFromString(val);
                            SyncButtonText();
                        };
                    }
                }
            }

            public UIDynamicButton button {
                get {
                    return _button;
                }
                set {
                    _button = value;
                    SetUpButton();
                }
            }

            public float height {
                get {
                    return _height;
                }
                set {
                    _height = value;
                    label.layoutElement.preferredHeight = _height;
                    if(button != null) {
                        button.height = _height;
                    }
                }
            }
        }


        /* 
            Random Utilities
        */
        public static string GetPathToObject(this MVRScript script, GameObject go){
            if (go.transform.parent == null || go.transform.parent == script.containingAtom.gameObject.transform)
                return "/" + go.name;
            return GetPathToObject(script, go.transform.parent.gameObject) + "/" + go.name;
        }
    }
}